"""Timeseer Client provides convenient remote access to Timeseer.

Data, metadata and event frames are exposed as Python objects."""

from kukur import DataType, Dictionary, InterpolationType, Metadata, ProcessType, SeriesSelector

from .base import AugmentationStrategy, TimeseerClientException
from .client import Client, filter_event_frames, filter_series

__all__ = [
    'AugmentationStrategy',
    'Client',
    'DataType',
    'Dictionary',
    'InterpolationType',
    'Metadata',
    'ProcessType',
    'SeriesSelector',
    'TimeseerClientException',
    'filter_event_frames',
    'filter_series',
]



"""The main objects in Timeseer client."""

from enum import Enum


class AugmentationStrategy(Enum):
    """AugmentationStrategy indicates the strategy for the data when filtering event frames."""
    REMOVE = 'remove values'
    HOLD_LAST = 'hold last value'
    LINEAR_INTERPOLATION = 'linear interpolation'
    KNN_IMPUTATION = 'knn imputation'


class TimeseerClientException(Exception):
    """Base class for Timeseer client exceptions.

    Use this to catch any exception that originates in the client."""


class AugmentationException(TimeseerClientException):
    """Exception raised when augmentation strategy fails."""



"""Filter event frames in Arrow tables or merge Arrow-backed event frame lists."""

from typing import List
from datetime import datetime

import pyarrow as pa
import pyarrow.compute as pc

from timeseer_client.base import AugmentationException
from .base import AugmentationStrategy


def filter_series(
    series: pa.Table,
    event_frames: pa.Table,
    augmentation_strategy: AugmentationStrategy,
) -> pa.Table:
    """Filter the even_frames out of the series based on the strategy."""

    if augmentation_strategy == AugmentationStrategy.KNN_IMPUTATION:
        raise AugmentationException()

    total = pa.array([True] * series.num_rows)

    for index, frame in enumerate(event_frames['start_date']):
        # pylint: disable=no-member
        on_or_after = pc.less(series['ts'], frame)
        before = pc.greater(series['ts'], event_frames['end_date'][index])
        out_event_frame = pc.or_(on_or_after, before)
        total = pc.and_(total, out_event_frame)

    if augmentation_strategy == AugmentationStrategy.HOLD_LAST:
        return _hold_last_value(series, total)

    if augmentation_strategy == AugmentationStrategy.LINEAR_INTERPOLATION:
        return _linear_interpolation(series, total)

    return series.filter(total)


def filter_event_frames(event_frames: pa.Table, start_date: datetime, end_date: datetime) -> pa.Table:
    """Restrict time range to filter event frames."""
    if event_frames.num_rows == 0:
        return event_frames

    new_start_dates = []
    new_end_dates = []
    total: List[bool] = [True] * event_frames.num_rows
    for index, start_date_frame in enumerate(event_frames['start_date']):
        end_date_frame = event_frames['end_date'][index]
        # pylint: disable=no-member
        out_start = pc.greater_equal(start_date_frame, pa.scalar(end_date))
        out_end = pc.less_equal(end_date_frame, pa.scalar(start_date))
        new_start_date = start_date_frame.as_py()
        new_end_date = end_date_frame.as_py()
        if out_start.as_py() or out_end.as_py():
            total[index] = False
        if pc.less(start_date_frame, pa.scalar(start_date)).as_py():
            new_start_date = start_date
        if pc.greater(end_date_frame, pa.scalar(end_date)).as_py():
            new_end_date = end_date
        new_start_dates.append(new_start_date)
        new_end_dates.append(new_end_date)
    new_table = event_frames.set_column(
                    0,
                    'start_date',
                    pa.array(new_start_dates, type=pa.timestamp('us', tz='UTC'))
                )
    new_table = new_table.set_column(
                    1,
                    'end_date',
                    pa.array(new_end_dates, type=pa.timestamp('us', tz='UTC'))
                )
    return new_table.filter(total)


def _hold_last_value(series: pa.Table, out_of_event_frames: pa.array) -> pa.Table:
    new_values = []
    new_ts = []
    last_acceptable = None
    for index, out_of_frame in enumerate(out_of_event_frames.to_pylist()):
        new_value = series['value'][index].as_py()

        if out_of_frame:
            last_acceptable = series['value'][index].as_py()
        else:
            if last_acceptable is None:
                continue
            new_value = last_acceptable

        new_values.append(new_value)
        new_ts.append(series['ts'][index].as_py())

    filtered_series = pa.Table.from_arrays([pa.array(new_ts), pa.array(new_values)], ['ts', 'value'])
    return filtered_series


def _linear_interpolation(series: pa.Table, out_of_event_frames: pa.array) -> pa.Table:
    new_values = []
    new_ts = []
    series_to_filter_count = 0
    last_acceptable = None
    for index, out_of_frame in enumerate(out_of_event_frames.to_pylist()):
        if out_of_frame:
            if series_to_filter_count != 0 and last_acceptable is not None:

                for element in range(series_to_filter_count):
                    numerator = series['ts'][index - (series_to_filter_count - element)].as_py() - \
                                    series['ts'][index - (series_to_filter_count + 1)].as_py()
                    denominator = series['ts'][index].as_py() \
                        - series['ts'][index - (series_to_filter_count + 1)].as_py()
                    time_interpolation = numerator / denominator
                    new_value = last_acceptable * (1 - time_interpolation) + \
                        series['value'][index].as_py() * time_interpolation
                    new_values.append(new_value)
                    new_ts.append(series['ts'][index - (series_to_filter_count - element)].as_py())

            last_acceptable = series['value'][index].as_py()
            new_ts.append(series['ts'][index].as_py())
            new_values.append(last_acceptable)
            series_to_filter_count = 0
        else:
            series_to_filter_count = series_to_filter_count + 1
    return pa.Table.from_arrays([pa.array(new_ts), pa.array(new_values)], ['ts', 'value'])



"""Timeseer Client allows querying of data and metadata."""

import json

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional

from kukur import InterpolationType, SeriesSelector
from kukur.client import Client as KukurClient

import pyarrow.flight as fl
try:
    import pandas as pd
    import timeseer_client.filters_pandas as filters_pandas
    HAS_PANDAS = True
except Exception:  # pylint: disable=broad-except
    HAS_PANDAS = False

import timeseer_client.filters_arrow as filters_arrow
from .base import AugmentationStrategy


class Client(KukurClient):
    """Client connects to Timeseer using Arrow Flight."""

    def get_event_frames(
        self,
        start_date: datetime = None,
        end_date: datetime = None,
        frame_type: str = None,
        selector: SeriesSelector = None,
    ):
        """Get all event frames matching the given criteria.

        Args:
            start_date: the start date of the range to find overlapping event frames in. Defaults to one year ago.
            end_date: the end date of the range to find overlapping event frames in. Defaults to now.
            frame_type: the type of event frames to search for. Finds all types when empty.
            selector: the time series source or time series to which the event frames are linked.
                Matches all by default.

        Returns::
            A pyarrow Table with 5 columns.
            The first column ('start_date') contains the start date.
            The second column ('end_date') contains the end date.
            The third column ('type') contains the type of the returned event frame as a string.
            Columns 4 ('series_source') and 5 ('series_name') contain the source and name of the series.
        """
        if start_date is None or end_date is None:
            now = datetime.utcnow().replace(tzinfo=timezone(timedelta(0)))
            if start_date is None:
                start_date = now.replace(year=now.year-1)
            if end_date is None:
                end_date = now

        query: Dict[str, Any] = {
            'query': 'get_event_frames',
            'start_date': start_date.isoformat(),
            'end_date': end_date.isoformat(),
        }

        if frame_type is not None:
            query['type'] = frame_type
        if selector is not None:
            query['selector'] = {
                'source': selector.source,
            }
            if selector.name is not None:
                query['selector']['name'] = selector.name

        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def get_data_quality_score_data_sources(self, source_names: List[str]):
        """Get the data quality score of a data source.

        Args:
            source_names: A list of time series sources

        Returns::
            A data quality score of every given source in percentage.
        """
        body = {
            'source_names': source_names,
        }
        results = list(self._get_client().do_action(("get_data_quality_score_data_sources", json.dumps(body).encode())))
        return json.loads(results[0].body.to_pybytes())

    def get_kpi_scores(
        self,
        source_name: str,
    ):
        """Get the kpi scores of a data source.

        Args:
            source_name: The time series source

        Returns::
            The score per KPI for the source in percentage.
        """
        body = {
            'source_name': source_name,
        }

        results = list(self._get_client().do_action(("get_kpi_scores", json.dumps(body).encode())))
        return json.loads(results[0].body.to_pybytes())

    def get_data_quality_score_project(self, project_name: str,):
        """Get the data quality score of a data source.

        Args:
            project_name: The name of the project in Timeseer.

        Returns::
            The data quality score of the project in percentage.
        """
        body = {
            'project_name': project_name,
        }
        results = list(self._get_client().do_action(("get_data_quality_score_project", json.dumps(body).encode())))
        return json.loads(results[0].body.to_pybytes())


def filter_series(
        series,
        event_frames,
        augmentation_strategy: AugmentationStrategy = AugmentationStrategy.REMOVE,
        interpolation_type: Optional[InterpolationType] = None,
        context: Optional[List[pd.DataFrame]] = None
):
    """Filter the time series in the time periods given by event_frames

        Args:
            series: a pyarrow Table or a pandas DataFrame with time series date
                Two columns are present: 'ts' and 'value'.
                A pandas DataFrame can contain a DatetimeIndex instead of the 'ts' column.
            event_frames: pyarrow Table or a pandas DataFrame with event frames.
                Three columns need to be present: 'type', 'start_date' and 'end_date'.
            augmentation_strategy: An enum to define which strategy to use when filtering.
                'REMOVE' (the default) removes the values, 'HOLD_LAST' keeps the last acceptable value and
                'LINEAR_INTERPOLATION' interpolates the last acceptable value and the next acceptable value.
                If no acceptable value exists, they are removed. 'KNN_IMPUTATION' uses context to find the
                k-nearest neighbors and takes the average.
            interpolation_type: Enum to define the interpolation type. 'LINEAR' or 'STEPPED'.
                Only linear interpolation types can be linearly interpolated in the augmentation strategy.
            context: A list of pd.DataFrame used for 'KNN_IMPUTATION' to find nearest neighbors
        Returns:
            A filtered pyarrow Table or a pandas DataFrame with 2 columns: 'ts' and 'value'.
            In case the pandas DataFrame provided in 'series' has a DatetimeIndex,
                the 'ts' column will not be present, but the DataFrame will have a DateTimeIndex.
    """
    if context is None:
        context = []

    if augmentation_strategy == AugmentationStrategy.LINEAR_INTERPOLATION and \
       interpolation_type != InterpolationType.LINEAR:
        augmentation_strategy = AugmentationStrategy.HOLD_LAST

    if len(series) == 0:
        return series

    if(HAS_PANDAS and isinstance(series, pd.DataFrame) and isinstance(event_frames, pd.DataFrame)):
        return filters_pandas.filter_series(series, event_frames, augmentation_strategy, context)
    return filters_arrow.filter_series(series, event_frames, augmentation_strategy)


def filter_event_frames(
        event_frames,
        start_date: datetime,
        end_date: datetime
):
    """Restrict the event frames to the given time range.

        Args:
            event_frames: a pyarrow Table or a pandas DataFrame with event frames.
            start_date: the start date of the range to filter event_frames.
            end_date: the end date of the range to filter event_frames.

        Returns::
            A filtered pyarrow Table or a pandas DataFrame with 5 columns.
            The first column ('start_date') contains the 'start_date' and 'end_date'.
            The second column ('end_date') contains the 'end_date'.
            The third column ('type') contains the type of the returned event frame as a string.
            Columns 4 ('series_source') and 5 ('series_name') contain the source and name of the series.
    """
    if(HAS_PANDAS and isinstance(event_frames, pd.DataFrame)):
        return filters_pandas.filter_event_frames(event_frames, start_date, end_date)
    return filters_arrow.filter_event_frames(event_frames, start_date, end_date)
